"""
This file contains all the constants used throughout
the Xiangqi environment.
"""

""" PATHS """
PATH_TO_SOUNDS = "sounds/"
PATH_TO_BOARD = "images/board/"
PATH_TO_BLACK = "images/black_pieces/"
PATH_TO_RED = "images/red_pieces/"

""" PYGAME """
WINDOW_WIDTH = 521
WINDOW_HEIGHT = 800
FPS = 20
COUNT = 10

""" POINTS """
PIECE_POINTS = [
    0.,                     # EMPTY: No point for empty grid
    float('inf'),           # GENERAL: Priceless for the general
    2., 2.,                 # ADVISOR: 2.0 points
    2., 2.,                 # ELEPHANT: 2.0 points
    4., 4.,                 # HORSE: 4.0 points
    9., 9.,                 # CHARIOT: 9.0 points
    4.5, 4.5,               # CANNON: 4.5 points
    1., 1., 1., 1., 1.,     # SOLDIER: 1.0 point (2.0 after crossing river)
]

ILLEGAL_MOVE = -10.
WIN = PIECE_POINTS[1]
LOSE = -PIECE_POINTS[1]

""" PIECE """
ALLY = 1
ENEMY = -1

PIECE_CNT = 16              # Total number of pieces in each side

# Piece IDs
EMPTY = 0
GENERAL = 1
ADVISOR_1 = 2
ADVISOR_2 = 3
ELEPHANT_1 = 4
ELEPHANT_2 = 5
HORSE_1 = 6
HORSE_2 = 7
CHARIOT_1 = 8
CHARIOT_2 = 9
CANNON_1 = 10
CANNON_2 = 11
SOLDIER_1 = 12
SOLDIER_2 = 13
SOLDIER_3 = 14
SOLDIER_4 = 15
SOLDIER_5 = 16
# TODO: Consolidate this list with the ID to
# prevent problems with synchronization.
PIECE_ID_TO_NAME = [
    "EMPTY", "GENERAL", "ADVISOR_1", "ADVISOR_2",
    "ELEPHANT_1", "ELEPHANT_2", "HORSE_1", "HORSE_2",
    "CHARIOT_1", "CHARIOT_2", "CANNON_1", "CANNON_2",
    "SOLDIER_1", "SOLDIER_2", "SOLDIER_3", "SOLDIER_4", "SOLDIER_5",
]

# Piece Movement Offsets
ORTHOGONAL = [(-1, 0), (0, 1), (1, 0), (0, -1)]
DIAGONAL = [(-1, 1), (1, 1), (1, -1), (-1, -1)]
ELEPHANT_MOVE = [(-2, 2), (2, 2), (2, -2), (-2, -2)]
HORSE_MOVE = [
    [(-1, 0), (-1, -1)], [(-1, 0), (-1, 1)],
    [(0, 1), (-1, 1)], [(0, 1), (1, 1)],
    [(1, 0), (1, 1)], [(1, 0), (1, -1)],
    [(0, -1), (1, -1)], [(0, -1), (-1, -1)]
]

# Piece States
RED = 0
BLACK = 1
DEAD = 0
ALIVE = 1

# Piece Size
PIECE_WIDTH = 58
PIECE_HEIGHT = 58
MINI_PIECE_WIDTH = 29
MINI_PIECE_HEIGHT = 29

""" BOARD """
# Board Size
BOARD_WIDTH = 521
BOARD_HEIGHT = 577
BOARD_Y_OFFSET = (WINDOW_HEIGHT/2 - BOARD_HEIGHT/2)

# Board Dimension
BOARD_ROWS = 10
BOARD_COLS = 9

# Palace coordinates
PALACE_ENEMY_ROW = (0, 2)
PALACE_ALLY_ROW = (7, 9)
PALACE_COL = (3, 5)

# River line
RIVER_LOW = 4
RIVER_HIGH = 5

INITIAL_BOARD = [
    [-9, -7, -5, -3, -1, -2, -4, -6, -8],
    [0, 0, 0, 0, 0, 0, 0, 0, 0],
    [0, -11, 0, 0, 0, 0, 0, -10, 0],
    [-16, 0, -15, 0, -14, 0, -13, 0, -12],
    [0, 0, 0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 0, 0, 0],
    [12, 0, 13, 0, 14, 0, 15, 0, 16],
    [0, 10, 0, 0, 0, 0, 0, 11, 0],
    [0, 0, 0, 0, 0, 0, 0, 0, 0],
    [8, 6, 4, 2, 1, 3, 5, 7, 9]
]

""" OTHER """
MAX_REP = 9         # number that is large enough to cover board width/height
TOTAL_POS = BOARD_ROWS * BOARD_COLS
MAX_PERPETUAL_JIANG = 4

""" Piece Coordinate Conversion """
COOR_DELTA = 57
COOR_OFFSET = 5



from gym_xiangqi.agents import RandomAgent
from gym_xiangqi.constants import ALLY, PIECE_ID_TO_NAME
from gym_xiangqi.utils import action_space_to_move

import gym
import time


def main():
    env = gym.make('gym_xiangqi:xiangqi-v0')
    env.render()
    agent = RandomAgent()

    done = False
    round = 0
    while not done:
        # Add a slight delay to properly visualize the game.
        time.sleep(1)

        action = agent.move(env)
        _, reward, done, _ = env.step(action)
        turn = "Ally" if env.turn == ALLY else "Enemy"
        move = action_space_to_move(action)
        piece = PIECE_ID_TO_NAME[move[0]]

        print(f"Round: {round}")
        print(f"{turn} made the move {piece} from {move[1]} to {move[2]}.")
        print(f"Reward: {reward}")
        print("================")

        round += 1
        env.render()
    env.close()


if __name__ == '__main__':
    main()



import os

import pkg_resources
import numpy as np
import pygame

from gym_xiangqi.utils import move_to_action_space, is_ally
from gym_xiangqi.constants import (
    ORTHOGONAL, DIAGONAL, ELEPHANT_MOVE, HORSE_MOVE,    # piece moves
    BOARD_ROWS, BOARD_COLS,                             # board specs
    PALACE_ALLY_ROW, PALACE_ENEMY_ROW, PALACE_COL,      # palace bound
    RIVER_LOW, RIVER_HIGH,                              # river bound
    MAX_REP,                                            # repetition bound
    BLACK, ALIVE, ALLY, ENEMY,                          # piece states
    COOR_DELTA, COOR_OFFSET,                            # board coordinate
    PIECE_WIDTH, PIECE_HEIGHT,                          # piece sizes
    MINI_PIECE_WIDTH, MINI_PIECE_HEIGHT,                # mini piece sizes
    PATH_TO_BLACK, PATH_TO_RED,                         # file paths to pieces
    EMPTY, GENERAL,                                     # piece IDs
    BOARD_Y_OFFSET                                      # board y offset
)


class Piece:
    """
    A base class for all Xiangqi pieces

    All pieces have the following:

        Attributes:
        - color: red or black
        - position: (row, column) coordinate
        - state: alive or dead (in game or out of game)
        - image: PyGame image object used when rendering

        Methods:
        - move(self): make allowed movements
    """

    def __init__(self, color, row, col):
        self.color = color
        self.row = row
        self.col = col
        self.state = ALIVE
        self.legal_moves = None
        self.piece_width = PIECE_WIDTH
        self.piece_height = PIECE_HEIGHT
        self.mini_piece_width = MINI_PIECE_WIDTH
        self.mini_piece_height = MINI_PIECE_HEIGHT
        self.basic_image = None
        self.select_image = None
        self.mini_image = None
        self.move_sound = None

    def move(self, new_row, new_col):
        """
        Take one move among given piece's allowed moves
        Update piece's coordinates internally
        """
        if self.move_sound is not None:
            self.move_sound.play()
        self.row = new_row
        self.col = new_col

    def get_pygame_coor(self):
        x = self.col*COOR_DELTA + COOR_OFFSET
        y = self.row*COOR_DELTA + COOR_OFFSET + BOARD_Y_OFFSET
        return (x, y)

    def load_image(self, filename: str, piece_width, piece_height):
        if self.color == BLACK:
            file_path = PATH_TO_BLACK
        else:
            file_path = PATH_TO_RED

        target_file = os.path.join(file_path, filename)
        target_file = pkg_resources.resource_filename(__name__, target_file)
        image = pygame.image.load(target_file).convert_alpha()
        image = pygame.transform.scale(
            image, (piece_width, piece_height)
        )
        return image

    def set_basic_image(self):
        filename = self.name + ".png"
        self.basic_image = (self.load_image(filename,
                            self.piece_width, self.piece_height))

    def set_select_image(self):
        filename = self.name + "_S.png"
        self.select_image = (self.load_image(filename,
                             self.piece_width, self.piece_height))

    def set_mini_image(self):
        filename = self.name + ".png"
        self.mini_image = (self.load_image(filename,
                           self.mini_piece_width, self.mini_piece_height))

    def is_alive(self):
        return self.state

    # getters
    @property
    def coor(self):
        return (self.col, self.row)


def check_action(piece_id, orig_pos, cur_pos,
                 repeat, offset, i, state, actions):
    """
    This is general searching procedure. Given the following parameters,
    repeatedly search in the same direction until either end of the board
    or another piece is blocking.

    Parameters:
        piece_id (int): piece ID
        orig_pos (tuple(int)): original coordinate of the piece
        cur_pos (tuple(int)): current position in evaluation
        repeat (int): number of repetitions to perform this procedure
        offset (tuple(int)): coordinate offset towards current direction
        i (int): current iteration number
        state (numpy.ndarray): current environment state
        actions (numpy.ndarray): pool of possible actions
    return:
        Number of times repeated; This is used to find out the farthest
        possible position used for other conditional check.
    """
    r = cur_pos[0]
    c = cur_pos[1]

    if not is_ally(piece_id):
        sign = ENEMY
        piece_id *= ENEMY
    else:
        sign = ALLY

    for i in range(repeat):
        rb = 0 <= r < BOARD_ROWS
        cb = 0 <= c < BOARD_COLS

        if not rb or not cb:
            return i

        # if ally piece is located, can't go further
        if state[r][c] * sign > 0:
            break

        if check_flying_general(state, sign, piece_id, orig_pos, (r, c)):
            return 0

        action_idx = move_to_action_space(piece_id, orig_pos, (r, c))
        actions[action_idx] = 1

        if state[r][c] != 0:
            break

        r += offset[0]
        c += offset[1]

    return i + 1


def check_flying_general(state, side, piece_id, start, end):
    """
    Check if given input action results in flying general

    Parameters:
        state (np.array): 2D array representing current state
        side (int): -1 or 1 representing enemy or ally side
        piece_id (int): piece ID
        start (tuple(int)): current coordinate of given piece
        end (tuple(int)): destination coordinate of given piece
    Return (bool):
        indicates whether the action results in flying general or not
    """

    # simulate input action without altering current game state
    new_state = np.array(state)
    new_state[start[0]][start[1]] = EMPTY
    new_state[end[0]][end[1]] = piece_id * side

    enemy_gen_row = -1
    enemy_gen_col = -1
    ally_gen_row = -1
    ally_gen_col = -1

    for r in range(PALACE_ENEMY_ROW[0], PALACE_ENEMY_ROW[1]+1):
        for c in range(PALACE_COL[0], PALACE_COL[1]+1):
            if new_state[r][c] == GENERAL * ENEMY:
                enemy_gen_row = r
                enemy_gen_col = c

    for r in range(PALACE_ALLY_ROW[0], PALACE_ALLY_ROW[1] + 1):
        for c in range(PALACE_COL[0], PALACE_COL[1] + 1):
            if new_state[r][c] == GENERAL * ALLY:
                ally_gen_row = r
                ally_gen_col = c

    # check if they are in the same column
    if enemy_gen_col != ally_gen_col:
        return False

    # check if anything is in between the two generals
    for r in range(enemy_gen_row+1, ally_gen_row):
        if new_state[r][ally_gen_col] != EMPTY:
            return False
    return True


class General(Piece):
    """
    This piece is equivalent to King. It is called "Jiang" or "Shuai"
    meaning Governor (red) and General (black) respectively.
    - Only one piece exists in each side
    - Can only move 1 unit of space orthogonally within the special square area
    """

    def __init__(self, color, row, col):
        super(General, self).__init__(color, row, col)
        self.name = "GEN"

    def get_actions(self, piece_id, state, actions):
        """
        Finds legal moves for the General
        """
        if not is_ally(piece_id):
            low = PALACE_ENEMY_ROW[0]
            high = PALACE_ENEMY_ROW[1]
        else:
            low = PALACE_ALLY_ROW[0]
            high = PALACE_ALLY_ROW[1]

        for offset in ORTHOGONAL:
            next_pos = (self.row + offset[0], self.col + offset[1])

            # general must stay in the palace: row and column bound check
            rb = low <= next_pos[0] <= high
            cb = PALACE_COL[0] <= next_pos[1] <= PALACE_COL[1]
            if rb and cb:
                check_action(piece_id, (self.row, self.col), next_pos,
                             1, offset, 0, state, actions)


class Advisor(Piece):
    """
    This piece is called "Shi" meaning advisor/scholar.
    - 2 pieces exist in each side
    - Can only move 1 unit of space diagonally within the special square area
    """

    def __init__(self, color, row, col):
        super(Advisor, self).__init__(color, row, col)
        self.name = "ADV"

    def get_actions(self, piece_id, state, actions):
        """
        Finds legal moves for the Advisors
        """
        if not is_ally(piece_id):
            low = PALACE_ENEMY_ROW[0]
            high = PALACE_ENEMY_ROW[1]
        else:
            low = PALACE_ALLY_ROW[0]
            high = PALACE_ALLY_ROW[1]

        for offset in DIAGONAL:
            next_pos = (self.row + offset[0], self.col + offset[1])

            # advisor must stay in the palace: row and column bound check
            rb = low <= next_pos[0] <= high
            cb = PALACE_COL[0] <= next_pos[1] <= PALACE_COL[1]
            if rb and cb:
                check_action(piece_id, (self.row, self.col), next_pos,
                             1, offset, 0, state, actions)


class Elephant(Piece):
    """
    It is called "Shiang" meaning minister (red) and elephant (black)
    This piece is similar to Bishop.
    - 2 pieces exist in each side
    - This piece cannot cross the river
    - Moves 2 unit of space diagonally
    """

    def __init__(self, color, row, col):
        super(Elephant, self).__init__(color, row, col)
        self.name = "ELE"

    def get_actions(self, piece_id, state, actions):
        """
        Finds legal moves for the Elephants
        """
        if not is_ally(piece_id):
            low = 0
            high = RIVER_LOW
        else:
            low = RIVER_HIGH
            high = BOARD_ROWS - 1

        for offset in ELEPHANT_MOVE:
            next_pos = (self.row + offset[0], self.col + offset[1])

            # bound check: must not cross the river
            rb = low <= next_pos[0] <= high
            cb = 0 <= next_pos[1] < BOARD_COLS
            if not rb or not cb:
                continue

            # must be not blocked
            block_r = self.row + offset[0] // 2
            block_c = self.col + offset[1] // 2
            if state[block_r][block_c] != 0:
                continue

            check_action(piece_id, (self.row, self.col), next_pos,
                         1, offset, 0, state, actions)


class Horse(Piece):
    """
    This piece is almost identical to knight in Chess.
    It is called "Ma" meaning horse.
    - 2 Pieces in each side.
    - Moves just like knights in chess. (The "L" shape move)
    - Cannot jump over pieces unlike Knights in Chess
    """

    def __init__(self, color, row, col):
        super(Horse, self).__init__(color, row, col)
        self.name = "HRS"

    def get_actions(self, piece_id, state, actions):
        """
        Finds legal moves for the Horses
        """
        # horse moves consist of 2 separate moves:
        # 1. along the line up or down or left or right
        # 2. diagonally left or right along the same direction
        for first_move, second_move in HORSE_MOVE:
            next_r = self.row + first_move[0]
            next_c = self.col + first_move[1]

            # bound check
            rb = 0 <= next_r < BOARD_ROWS
            cb = 0 <= next_c < BOARD_COLS

            if not rb or not cb:
                continue

            # check for any blocking piece
            if state[next_r][next_c] != 0:
                continue

            next_pos = (next_r + second_move[0], next_c + second_move[1])

            # no need to recurse on next moves; (0, 0) just a placeholder
            check_action(piece_id, (self.row, self.col), next_pos,
                         1, (0, 0), 0, state, actions)


class Chariot(Piece):
    """
    It is called "Chuh" meaning chariot
    2 pieces in each side.
    This piece is identical to Rook in Chess.
    Moves just like a Rook in chess
    - As many as you want horizontally or vertically.
    """

    def __init__(self, color, row, col):
        super(Chariot, self).__init__(color, row, col)
        self.name = "CHR"

    def get_actions(self, piece_id, state, actions):
        """
        Find legal moves for the Chariots
        """
        for offset in ORTHOGONAL:
            next_pos = (self.row + offset[0], self.col + offset[1])
            # No need to check for repetition; check as far as possible
            check_action(piece_id, (self.row, self.col), next_pos,
                         MAX_REP, offset, 0, state, actions)


class Cannon(Piece):
    """
    It is called "Pao" meaning cannon or catapult.
    2 pieces in each side.
    It moves similar to a Rook. The difference is, it has to jump over
    ONE piece (enemy or foe) to capture enemy piece.
    """

    def __init__(self, color, row, col):
        super(Cannon, self).__init__(color, row, col)
        self.name = "CAN"

    def get_actions(self, piece_id, state, actions):
        """
        Find legal moves for the Cannons
        """
        if not is_ally(piece_id):
            sign = ENEMY
        else:
            sign = ALLY

        for offset in ORTHOGONAL:
            # moving positions
            next_pos = (self.row + offset[0], self.col + offset[1])
            reps = check_action(piece_id, (self.row, self.col), next_pos,
                                MAX_REP, offset, 0, state, actions)

            # mark the farthest position invalid if it is an enemy
            last_r = self.row + offset[0] * reps
            last_c = self.col + offset[1] * reps

            if state[last_r][last_c] * sign < 0:
                action_idx = move_to_action_space(piece_id * sign,
                                                  (self.row, self.col),
                                                  (last_r, last_c))
                actions[action_idx] = 0

            # attacking positions
            next_r = self.row + offset[0] * (reps + 1)
            next_c = self.col + offset[1] * (reps + 1)

            while True:
                rb = 0 <= next_r < BOARD_ROWS
                cb = 0 <= next_c < BOARD_COLS

                if not rb or not cb:
                    break

                if state[next_r][next_c] * sign > 0:
                    break
                elif state[next_r][next_c] * sign < 0:
                    if check_flying_general(state, sign, piece_id,
                                            (self.row, self.col),
                                            (next_r, next_c)):
                        break

                    action_idx = move_to_action_space(
                        piece_id * sign, (self.row, self.col), (next_r, next_c)
                    )
                    actions[action_idx] = 1
                    break

                next_r += offset[0]
                next_c += offset[1]


class Soldier(Piece):
    """
    It is called "Ping" (red) and "Tsuh" (black) meaning a foot soldier.
    5 pieces in each side.
    This is equivalent to Pawn in chess.
    Moves 1 unit of space forward
    When it crosses the river, it gets options to move left or right as well.
    """

    def __init__(self, color, row, col):
        super(Soldier, self).__init__(color, row, col)
        self.name = "SOL"

    def get_actions(self, piece_id, state, actions):
        """
        Find legal moves for the soldiers
        """
        # ORTHOGONAL contains 4 moves in clock-wise [UP, RIGHT, DOWN, LEFT]
        if not is_ally(piece_id):   # enemy side is always at the top half
            low = RIVER_HIGH
            high = BOARD_ROWS - 1
            moves = [2]             # therefore enemy soldiers move downwards
        else:                       # ally side is always at the bottom half
            low = 0
            high = RIVER_LOW
            moves = [0]             # therefore ally soldiers move upwards

        # low and high are set to be after-river row ranges
        if low <= self.row <= high:     # After crossing the river,
            moves.append(1)             # can move right
            moves.append(3)             # can move left

        for i in moves:
            offset = ORTHOGONAL[i]
            next_pos = (self.row + offset[0], self.col + offset[1])
            check_action(piece_id, (self.row, self.col), next_pos,
                         1, offset, 0, state, actions)



import unittest

from gym_xiangqi.utils import move_to_action_space, action_space_to_move


class TestXiangQiUtils(unittest.TestCase):

    def test_move_to_action_space(self):
        i = 0
        for pid in range(1, 17):
            for r1 in range(10):
                for c1 in range(9):
                    for r2 in range(10):
                        for c2 in range(9):
                            action_idx = move_to_action_space(
                                pid, (r1, c1), (r2, c2)
                            )
                            self.assertEqual(action_idx, i)
                            i += 1

    def test_action_space_to_move(self):
        def info_gen():
            for piece in range(1, 17):
                for r1 in range(10):
                    for c1 in range(9):
                        for r2 in range(10):
                            for c2 in range(9):
                                yield piece, (r1, c1), (r2, c2)

        gen = info_gen()
        n = pow(10 * 9, 2) * 16

        for i in range(n):
            pid, start, end = action_space_to_move(i)
            ans_pid, ans_start, ans_end = next(gen)
            self.assertEqual(pid, ans_pid)
            self.assertEqual(start[0], ans_start[0])
            self.assertEqual(start[1], ans_start[1])
            self.assertEqual(end[0], ans_end[0])
            self.assertEqual(end[1], ans_end[1])


if __name__ == "__main__":
    unittest.main()



import gym
import pytest

from gym_xiangqi.agents import RandomAgent
from gym_xiangqi.constants import (
    EMPTY, ADVISOR_1
)


@pytest.fixture
def agent():
    """
    Create the agent which returns random moves.
    """
    return RandomAgent()


@pytest.fixture
def legal_env(mocker):
    """
    Create an environment where the game is
    still ongoing and there are possible actions.
    """
    env = gym.make('gym_xiangqi:xiangqi-v0')
    env.ally_actions.fill(0)
    # id: 2 (ADVISOR_1), start: [9, 3], end: [8, 4]
    env.ally_actions[15736] = 1
    return env


def test_make_legal_move(agent, legal_env):
    action = agent.move(legal_env)
    assert legal_env.action_space.contains(action)
    assert legal_env.ally_actions[action] == 1

    state, reward, done, _ = legal_env.step(action)
    assert state[9][3] == EMPTY
    assert state[8][4] == ADVISOR_1
    assert reward == 0
    assert done is False



from gym_xiangqi.agents import RandomAgent

import gym

MAX_ROUNDS = 500


def test_random_agent_play_itself():
    """
    Test integration between agent and env by
    playing a game against itself.
    """
    env = gym.make('gym_xiangqi:xiangqi-v0')
    agent = RandomAgent()

    done = False
    round_count = 0
    while not done:
        action = agent.move(env)
        _, _, done, _ = env.step(action)
        round_count += 1
        if round_count > MAX_ROUNDS:
            break
    env.close()



import gym
from gym import spaces
from gym.utils import seeding
import numpy as np

from gym_xiangqi.xiangqi_game import XiangQiGame
from gym_xiangqi.utils import (
    action_space_to_move,
    move_to_action_space,
    is_ally
)
from gym_xiangqi.piece import (
    General, Advisor, Elephant, Horse, Chariot, Cannon, Soldier
)
from gym_xiangqi.constants import (
    INITIAL_BOARD,
    BOARD_ROWS, BOARD_COLS,
    TOTAL_POS, PIECE_CNT,
    RED, BLACK, ALIVE, DEAD,
    ILLEGAL_MOVE, PIECE_POINTS, LOSE,
    ALLY, ENEMY, EMPTY, GENERAL,
    MAX_PERPETUAL_JIANG,
)


class XiangQiEnv(gym.Env):
    """
    This is Xiangqi (Chinese chess) game implemented as reinforcement
    learning environment using OpenAI Gym framework. Xiangqi is played
    on a board of 10 rows and 9 columns with 16 pieces on each side (7
    unique pieces called General, Advisor, Elephant, Horse, Chariot,
    Cannon and Soldier.

    Starting State:
    The initial board state with pieces laid out in correct position.
    Reference the README for initial board illustration.

    Episode Termination:
    Either the red or black runs out of moves or also known as the
    general is captured. Reference the README for more details.

    Attributes:
        observation_space (gym.spaces.Box(10, 9)):
            The observation space is the state of the board and pieces.
            Each item in the space corresponds to a single coordinate on
            the board with the value range from -16 to 16. Each piece is
            encoded as an integer in that range. Negative integers are enemy
            pieces and positive integers are ally pieces.

        action_space (gym.spaces.Discrete(16 * 10 * 9 * 10 * 9)):
            The action space is an aggregation of all possible moves even
            including illegal moves. Each space encodes 3 information: which
            piece, from where, and to where. From 16 * 10 * 9 * 10 * 9, 16 is
            the number of pieces and 10 * 9 is all possible grid positions on
            the board. The first 10 * 9 represents the start position and the
            second half represents the end position which is the position the
            piece wants to move to.

            In addition to this, the environment will calculate legal and
            illegal moves within the action space to penalize an agent trying
            to perform illegal moves and to correctly implement Xiangqi rules.

        ally_color (int):
            Current environment's ally color
            RED = 0 and BLACK = 1

        enemy_color (int):
            Current environment's enemy color
            RED = 0 and BLACK = 1

        turn (int):
            Current player that is playing
            ALLY = 0 and ENEMY = 1

        done (bool):
            flag to indicate current game termination

        state (np.array):
            2 dimensional numpy array representing current board state

        ally_actions (np.array):
            1 dimensional numpy array indicating legal and illegal actions
            among all ally's action space
            Possible values of the array are 0 and 1 indicating legal and
            illegal actions respectively

        enemy_actions (np.array):
            1 dimensional numpy array indicating legal and illegal actions
            among all enemy's action space
            Possible values of the array are 0 and 1 indicating legal and
            illegal actions respectively

        ally_piece (list):
            List of all ally piece objects

        enemy_piece (list):
            List of all enemy piece objects
    """
    metadata = {'render.modes': ['human']}

    id_to_class = [
        None,
        General,
        Advisor, Advisor,
        Elephant, Elephant,
        Horse, Horse,
        Chariot, Chariot,
        Cannon, Cannon,
        Soldier, Soldier, Soldier, Soldier, Soldier
    ]

    def __init__(self, ally_color=RED):
        self._ally_color = ally_color
        if ally_color == RED:
            self._enemy_color = BLACK
            self._turn = ALLY
        else:
            self._enemy_color = RED
            self._turn = ENEMY

        # epoch termination flag
        self._done = False
        self._done_warn = False

        # observation space: 10 x 9 space with pieces encoded as integers
        self.observation_space = spaces.Box(
            low=-PIECE_CNT,
            high=PIECE_CNT,
            shape=(BOARD_ROWS, BOARD_COLS),
            dtype=int
        )

        # action space: encodes start and target position and specific piece
        n = pow(TOTAL_POS, 2) * PIECE_CNT
        self.action_space = spaces.Discrete(n)

        # initial board state
        self._state = None
        self._state_hash = None

        # instantiate piece objects
        self._ally_piece = [None for _ in range(PIECE_CNT + 1)]
        self._enemy_piece = [None for _ in range(PIECE_CNT + 1)]

        # possible moves: binary list with same shape of action space
        #                 valid action will be represented as 1 else 0
        self._ally_actions = np.zeros((n, ))
        self._enemy_actions = np.zeros((n, ))

        # history of consecutive jiangs (will be used to ban perpetual check)
        self._ally_jiang_history = None
        self._enemy_jiang_history = None

        # initialize PyGame module
        self._game = XiangQiGame()

        # user movement information during user vs agent game mode
        self.user_move_info = None

        # reset all environment components to initial state
        self.reset()

    def step(self, action):
        """
        Run one turn of Xiangqi game (ally or enemy side plays a move)
        by processing given action based on current game turn owner

        Parameters:
            action (int): a valid action in Xiangqi action space

        Return:
            tuple: observation, reward, done, info

            observation (object): current game state of the environment

            reward (float): amount of reward returned after given action

            We apply points to every type of pieces following the most widely
            used standard.

            General: infinity

            Advisor: 2.0

            Elephant: 2.0

            Horse: 4.0

            Chariot: 9.0

            Cannon: 4.5

            Soldier: 1.0 (2.0 if it has crossed the river)

            done (bool): whether the episode has ended, in which case further
            step() calls will return undefined results

            info (dict): contains auxiliary diagnostic information (helpful for
            debugging, and sometimes learning)
        """
        # Validate action input
        error_msg = "%r (%s) invalid action" % (action, type(action))
        assert self.action_space.contains(action), error_msg

        # Validate that the environment wasn't changed between steps
        assert hash(str(self._state)) == self._state_hash, \
            "Error! Game state changed illegally!"

        # Warn the user for calling step() when current game has finished
        if self._done:
            if not self._done_warn:
                self._done_warn = True
                print(gym.utils.colorize(
                    "WARN: Environment should be reset with call to reset() "
                    "when the episode has terminated (i.e 'done == True')",
                    "yellow"
                ))
            return np.array(self._state), 0, self._done, {}

        # Prepare game state variables
        reward = 0.0

        if self._turn == ALLY:
            pieces = self._ally_piece
            possible_actions = self._ally_actions
            jiang_history = self._ally_jiang_history
        else:
            pieces = self._enemy_piece
            possible_actions = self.enemy_actions
            jiang_history = self._enemy_jiang_history

        # Check for illegal move, flying general, etc. and penalize the agent
        if possible_actions[action] == 0:
            return np.array(self._state), ILLEGAL_MOVE, False, {}

        # Check if opponent is in Jiang condition before processing given move
        pre_jiang_actions = self.check_jiang()

        # Move the piece if legal move is given
        piece, start, end = action_space_to_move(action)
        pieces[piece].move(*end)

        # Update observation space
        self._state[start[0]][start[1]] = EMPTY
        rm_piece_id = self._state[end[0]][end[1]]
        self._state[end[0]][end[1]] = piece * self._turn

        if rm_piece_id < 0:
            self._enemy_piece[-rm_piece_id].state = DEAD
        elif rm_piece_id > 0:
            self._ally_piece[rm_piece_id].state = DEAD

        # Reward based on removed piece
        reward += PIECE_POINTS[abs(rm_piece_id)]

        # End game if the General on either side has been attacked
        if abs(rm_piece_id) == GENERAL:
            self._done = True

        # Check for perpetual check/jiang
        post_jiang_actions = self.check_jiang()

        if post_jiang_actions:
            for jiang_action in post_jiang_actions:
                if jiang_action in pre_jiang_actions:
                    continue

                if jiang_action not in jiang_history:
                    jiang_history[jiang_action] = 0
                jiang_history[jiang_action] += 1

                if jiang_history[jiang_action] == MAX_PERPETUAL_JIANG:
                    self._done = True
                    return np.array(self._state), LOSE, self._done, {}
        else:       # reset history if jiang spree has stopped
            if self._turn == ALLY:
                self._ally_jiang_history = {}
            else:
                self._enemy_jiang_history = {}

        # self-play: agent switches turn between ally and enemy side
        self._turn *= -1     # ALLY (1) to ENEMY (-1) and vice versa
        self.get_possible_actions(self._turn)

        # Update state hash.
        self._state_hash = hash(str(self._state))

        return np.array(self._state), reward, self._done, {}

    def reset(self):
        """
        Reset all environment components to initial state

        Return:
            observation (object): the initial observation.
        """
        self._done = False
        self._state = np.array(INITIAL_BOARD)
        self.init_pieces()

        self._ally_jiang_history = {}
        self._enemy_jiang_history = {}

        if self._ally_color == RED:
            self._turn = ALLY
        else:
            self._turn = ENEMY

        self.get_possible_actions(self._turn)
        self._game.set_pieces(self._ally_piece, self._enemy_piece)
        self._state_hash = hash(str(self._state))

        return np.array(self._state)

    def render(self, mode='human'):
        """
        Render current game state with PyGame

        For more information refer to gym.Env.render() in OpenAI Gym
        repository.

        Parameters:
            mode (str): string to indicate render mode
        """
        if self._game.display_surf is None:
            self._game.on_init()
        if self._ally_piece[GENERAL].basic_image is None:
            self._game.on_init_pieces()
        self._game.render()

    def close(self):
        """
        Free up resources and gracefully exit the Xiangqi environment
        """
        if self._game:
            self._game.cleanup()

    def seed(self, seed=None):
        """
        Generate random seed value used to reproduce the current game

        Parameters:
            seed:
                User defined input seed. If this is `None`, then it is
                generated by this method
        """
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def step_user(self):
        """
        This method functions like the environment's step() method, but
        it is specifically for users when they are player of a Xiangqi
        game. The method first renders game GUI and listens to user inputs.
        Then, the user input, the piece movement, is converted into action
        space and passed to environment's step() method. The environment then
        is able to handle the input action just like it handles any actions
        from RL agents.

        Return:
            tuple:
            The return values are the same with step() method.
        """
        error_msg = "gym_xiangqi error: calling step_user with " \
                    "incorrect game turn (must be ally's turn)"
        assert self._turn == ALLY, error_msg

        for piece_id in range(1, PIECE_CNT+1):
            self.get_possible_actions_by_piece(piece_id)

        self._game.run()

        # Game terminated by window close button
        if self._game.quit:
            return self._state, 0, True, {"exit": True}

        # Retrieve user piece movement info
        piece_id = self._game.cur_selected_pid
        start = (self._ally_piece[piece_id].row,
                 self._ally_piece[piece_id].col)
        end = self._game.end_pos

        # Reset the variables
        self._game.cur_selected_pid = None
        self._game.end_pos = None

        # Save as instance variables for debugging
        self.user_move_info = (piece_id, start, end)

        # Process the piece movement in env
        player_action = move_to_action_space(piece_id, start, end)
        return self.step(player_action)

    def init_pieces(self):
        """
        Initialize and store all ally and enemy pieces
        """
        for r in range(BOARD_ROWS):
            for c in range(BOARD_COLS):
                piece_id = INITIAL_BOARD[r][c]
                init = self.id_to_class[abs(piece_id)]
                if piece_id < 0:
                    self._enemy_piece[-piece_id] = init(self._enemy_color,
                                                        r, c)
                elif piece_id > 0:
                    self._ally_piece[piece_id] = init(self._ally_color, r, c)

    def get_possible_actions(self, player):
        """
        Searches all valid actions each piece can perform

        Parameters:
            player (int): -1 for ENEMY 1 for ALLY
        """
        # Current piece set changes depending on whose turn it is
        if player == ALLY:
            piece_set = self._ally_piece
            possible_actions = self._ally_actions
        else:
            piece_set = self._enemy_piece
            possible_actions = self._enemy_actions

        # Clear previous turn's possible actions
        # TODO: Don't clear the entire list, but only the relevant actions.
        possible_actions.fill(0)

        # Get possible moves for every piece in the piece set
        for pid, piece_obj in enumerate(piece_set[1:], 1):
            if piece_obj.state == ALIVE:
                piece_obj.get_actions(pid * self._turn,
                                      self._state,
                                      possible_actions)

    def get_possible_actions_by_piece(self, piece_id):
        """
        Given a piece_id, returns only the possible actions that
        can be taken by the piece.

        Parameters:
            piece_id (int): Piece ID to filter possible actions.
        Return:
            actions that are can be taken by the piece.
        """
        if is_ally(piece_id):
            pieces = self._ally_piece
            possible_actions = self._ally_actions
        else:
            pieces = self._enemy_piece
            possible_actions = self._enemy_actions

        piece_id = abs(piece_id)

        # Calculate the starting and ending action index of the piece
        piece_action_id_start = (piece_id - 1) * pow(TOTAL_POS, 2)
        piece_action_id_end = piece_action_id_start + pow(TOTAL_POS, 2)

        # First, filter to obtain only legal actions
        legal_actions = np.where(possible_actions == 1)[0]

        # Second, filter the legal actions using the start and end action index
        legal_actions = legal_actions[legal_actions >= piece_action_id_start]
        legal_actions = legal_actions[legal_actions < piece_action_id_end]

        # Save the start and end coordinates in each piece object's legal_moves
        pieces[piece_id].legal_moves = [
            action_space_to_move(action)[1:] for action in legal_actions
        ]

    def check_jiang(self):
        """
        Check if the general is in threat (i.e. it is check or "jiang")
        by any of current player's pieces
        """
        # Get OPPONENT General
        if self._turn == ALLY:
            general = self._enemy_piece[GENERAL]
            actions = self._ally_actions
        else:
            general = self._ally_piece[GENERAL]
            actions = self._enemy_actions

        # Update current player's moves
        self.get_possible_actions(self._turn)

        # Iterate through possible moves of current player's pieces
        actions = np.where(actions == 1)[0]
        jiang_actions = []
        for action in actions:
            _, _, (target_r, target_c) = action_space_to_move(action)
            if target_r == general.row and target_c == general.col:
                jiang_actions.append(action)
        return jiang_actions

    @property
    def ally_color(self):
        return self._ally_color

    @property
    def enemy_color(self):
        return self._enemy_color

    @property
    def turn(self):
        return self._turn

    @property
    def state(self):
        return self._state

    @property
    def ally_piece(self):
        return self._ally_piece

    @property
    def enemy_piece(self):
        return self._enemy_piece

    @property
    def ally_actions(self):
        return self._ally_actions

    @property
    def enemy_actions(self):
        return self._enemy_actions

    @property
    def game(self):
        return self._game



from gym_xiangqi.constants import TOTAL_POS


def move_to_action_space(piece_id, start, end):
    """
    The action space is a 1D flat array. We can convert piece id,
    start position and end position to a corresponding index value
    in the action space.

    Parameters:
        piece_id (int): a piece ID integer
        start (tuple(int)): (row, col) start coordinate
        end (tuple(int)): (row, col) end coordinate
    return:
        Index within the self.possible_actions
    """
    piece_id_val = (piece_id - 1) * pow(TOTAL_POS, 2)
    start_val = (start[0] * 9 + start[1]) * TOTAL_POS
    end_val = end[0] * 9 + end[1]
    return piece_id_val + start_val + end_val


def action_space_to_move(action):
    """
    This is exact opposite of move_to_action_space() method.
    With index value, we can convert this back to piece id,
    start position and end position values.

    Parameters:
        action (int): index value within action space
    return:
        piece ID, start coordinate, end coordinate
    """
    piece_id, r = divmod(action, pow(TOTAL_POS, 2))
    start_val, end_val = divmod(r, TOTAL_POS)
    start = [0, 0]
    end = [0, 0]
    start[0], start[1] = divmod(start_val, 9)
    end[0], end[1] = divmod(end_val, 9)
    return piece_id + 1, start, end


def is_ally(piece_id):
    """
    Determines if given input piece_id is ally or enemy piece
    This function CANNOT guarantee if the piece is an enemy piece

    Return:
        True: given piece ID is an ally piece
        False: given piece ID is either an empty space or an enemy piece
    """
    return piece_id > 0



import time

import pygame

from gym_xiangqi.sound import Sound
from gym_xiangqi.board import Board
from gym_xiangqi.constants import (
    COOR_DELTA, COOR_OFFSET,      # variables for coordinate conversion
    DEAD,                         # dead state for piece object
    WINDOW_WIDTH, WINDOW_HEIGHT,  # window size for pygame display
    FPS,                          # fps for pygame while loop
    COUNT,                        # initial time for timer
    PIECE_CNT,                    # total number of pieces in each side
    BOARD_Y_OFFSET,               # board y offset
    BOARD_WIDTH, BOARD_HEIGHT,    # board width, height
    GENERAL,                      # Piece IDs
)


class XiangQiGame:
    """
    This class represents the Xiangqi game using PyGame.
    It is also a medium in which different components interact as a whole.
    For example, rendering the game graphics, taking user input from the game
    and providing information necessary for our XiangQiEnv class to function as
    reinforcement learning environment are all communicated and integrated
    through this class.
    """

    def __init__(self):
        # PyGame components
        self.running = True
        self.dim = (WINDOW_WIDTH, WINDOW_HEIGHT)
        self.display_surf = None
        self.sound = None
        self.ally_piece = None
        self.enemy_piece = None
        self.cur_selected = None
        self.ally_turn = True
        self.counter = COUNT
        self.ally_kills = []
        self.enemy_kills = []
        self.cur_selected_pid = 0
        self.end_pos = None
        self.bgm_switch = True
        self.quit = False
        self.compart_color = (153, 102, 75)

    def on_init(self):
        """
        Initialize/start the game with PyGame
        ex. pygame.init()
            (anything related to PyGame module that needs to be initialized)
        """
        pygame.init()
        self.display_surf = pygame.display.set_mode(
            self.dim,
            pygame.HWSURFACE | pygame.DOUBLEBUF
        )

        # init timer
        # self.init_timer() # hidden for now

        # set caption
        self.screen = pygame.display.set_mode(self.dim)
        pygame.display.set_caption("AI Xiangqi(Chinese Chess)")

        # init board
        self.board_background = self.init_board()

        # load game sound components
        self.init_sound("piece_move.wav", "bgm.wav")

    def set_pieces(self, ally_piece, enemy_piece):
        self.ally_piece = ally_piece
        self.enemy_piece = enemy_piece

    def on_init_pieces(self):
        # load piece images
        self.load_piece_images(self.ally_piece)
        self.load_piece_images(self.enemy_piece)

        # load move_sound and set it to piece objects
        for i in range(1, PIECE_CNT+1):
            self.ally_piece[i].move_sound = self.sound.piece_move
            self.enemy_piece[i].move_sound = self.sound.piece_move

    def init_board(self):
        '''
        Initializes Board() and load the board image
        '''
        # set board_background
        board = Board()
        board_image = pygame.transform.scale(
            board.board_background, (board.boardWidth, board.boardHeight)
        )
        return board_image

    def init_sound(self, piece_move, bgm):
        """
        Initialize game sound
        """
        self.sound = Sound(piece_move, bgm)

        # play bgm
        pygame.mixer.music.play(-1)

    def update_pos_next_moves(self):
        if self.cur_selected is None:
            return

        opacity = 128
        cur_sel_basic_img = self.cur_selected.basic_image.copy()
        cur_sel_basic_img.set_alpha(opacity)

        for _, (row, col) in self.cur_selected.legal_moves:
            pygame_y = row*COOR_DELTA + COOR_OFFSET + BOARD_Y_OFFSET
            pygame_x = col*COOR_DELTA + COOR_OFFSET
            self.screen.blit(cur_sel_basic_img, (pygame_x, pygame_y))

    def toggle_bgm(self):
        """
        This is a toggle switch for bgm.
        If self.bgm_switch is True, BGM plays, otherwise it stops.
        """
        self.bgm_switch = not self.bgm_switch
        if self.bgm_switch:
            pygame.mixer.music.play(-1)
        else:
            pygame.mixer.music.stop()

    def update_bgm_state(self):
        """
        Show whether the bgm is on or off
        """
        bgm_text = " BGM(B)"
        bgm_font = pygame.font.SysFont('bradleyhand', 20)
        bgm_state_font = pygame.font.SysFont('bradleyhand', 25)
        border_font = pygame.font.SysFont('bradleyhand', 30)

        if self.bgm_switch:
            color = (230, 100, 100)

            # Write ON in RED
            bgm_state_text = "ON"
            final_text = bgm_state_font.render(bgm_state_text, True, color)
            text_rect = final_text.get_rect(centerx=466, bottom=85)
            self.screen.blit(final_text, text_rect)
        else:
            color = (100, 100, 200)

            # Write ON in BLUE
            bgm_state_text = "OFF"
            final_text = bgm_state_font.render(bgm_state_text, True, color)
            text_rect = final_text.get_rect(centerx=466, bottom=85)
            self.screen.blit(final_text, text_rect)

        # Write BGM(B) in RED or BLUE
        final_text = bgm_font.render(bgm_text, True, color)
        text_rect = final_text.get_rect(centerx=466, bottom=50)
        self.screen.blit(final_text, text_rect)

        # Draw border line in RED or BLUE
        border_text = "|"
        final_text = border_font.render(border_text, True, color)
        text_rect = final_text.get_rect(centerx=WINDOW_WIDTH-100, bottom=60)
        self.screen.blit(final_text, text_rect)
        text_rect = final_text.get_rect(centerx=WINDOW_WIDTH-100, bottom=90)
        self.screen.blit(final_text, text_rect)

    def on_event(self, event):
        """
        This routine is triggered when some kind of user/game event is detected
        ex. when user closes the PyGame window
            (mostly any keyboard/mouse input)

        Parameter:
            event: PyGame event object that represents keyboard inputs, mouse
                   inputs and etc.
        """
        if event.type == pygame.QUIT:
            self.running = False
            self.quit = True

        elif event.type == pygame.KEYUP:
            if event.key == pygame.K_b:
                self.toggle_bgm()

        elif event.type == pygame.MOUSEBUTTONDOWN:
            # clicked: 1 not_clicked: 0
            left_clicked = pygame.mouse.get_pressed()[0]

            if left_clicked:
                # get clicked coordinate
                clicked_x, clicked_y = pygame.mouse.get_pos()
                clicked_coor = (clicked_x, clicked_y)

                # select any ally pieces that is in the clicked range
                self.find_target_piece(clicked_coor)

                # if another click is made while an ally piece is selected,
                # this must be a piece movement
                if self.cur_selected is not None:
                    # convert to real coordinate
                    real_clicked_coor = self.to_real_coor(clicked_coor)

                    # define the piece movement as start to end position
                    s = [self.cur_selected.row, self.cur_selected.col]
                    e = list(real_clicked_coor[::-1])

                    # validate the piece movement
                    is_legal = (s, e) in self.cur_selected.legal_moves

                    if is_legal:
                        self.end_pos = tuple(real_clicked_coor[::-1])

                        # reset counter after ally turn is over
                        self.counter = COUNT

                        # reset piece selection and end my turn
                        self.cur_selected = None
                        self.running = False
        """
        # timer decrement every second
        elif event.type == pygame.USEREVENT + 1:
            self.counter -= 1

            # timeout event
            if self.counter == 0:
                self.running = False
        """

    def on_update(self):
        pass

    def render(self):
        """
        Render current game state into graphics
        """
        # draw background & compartment lines
        self.draw_background()

        # self.update_timer() # hidden for now
        self.update_kills()
        self.update_bgm_state()
        self.screen.blit(self.board_background, (0, BOARD_Y_OFFSET))

        # update all cur positions of pieces
        for i in range(1, len(self.ally_piece)):
            if self.ally_piece[i].is_alive():
                self.screen.blit(self.ally_piece[i].basic_image,
                                 self.ally_piece[i].get_pygame_coor())
            if self.enemy_piece[i].is_alive():
                self.screen.blit(self.enemy_piece[i].basic_image,
                                 self.enemy_piece[i].get_pygame_coor())

        # if a piece is selected and alive, load select_image instead
        if self.cur_selected is not None and self.cur_selected.is_alive():
            self.screen.blit(self.cur_selected.select_image,
                             self.cur_selected.get_pygame_coor())

        self.update_pos_next_moves()
        self.render_kills()

        # draw all on screen
        pygame.display.update()

    def cleanup(self):
        """
        Free resources and exit the game and
        """
        pygame.quit()

    def reset(self):
        """
        Terminate current game and create a new game
        """
        pass

    def run(self):
        """
        Run the game until terminating condition is achieved
        """
        if self.display_surf is None:
            self.on_init()
        if self.ally_piece[GENERAL].basic_image is None:
            self.on_init_pieces()

        clock = pygame.time.Clock()
        self.running = True

        while self.running:
            clock.tick(FPS)
            for event in pygame.event.get():
                self.on_event(event)
            self.render()

    def draw_background(self):
        """
        This method draws the background
        and the compartment lines on the game screen.
        """
        # fill the background with white
        self.screen.fill((255, 255, 255))

        # horizontal compartment lines
        line_info = (0, 0, BOARD_WIDTH, 5)
        self.screen.fill(self.compart_color, line_info)
        line_info = (0, BOARD_Y_OFFSET-5, BOARD_WIDTH, BOARD_HEIGHT+10)
        self.screen.fill(self.compart_color, line_info)
        line_info = (0, WINDOW_HEIGHT-5, BOARD_WIDTH, 5)
        self.screen.fill(self.compart_color, line_info)

        # vertical compartment lines
        line_info = (0, 0, 10, WINDOW_HEIGHT)
        self.screen.fill(self.compart_color, line_info)
        line_info = (WINDOW_WIDTH-10, 0, 10, WINDOW_HEIGHT)
        self.screen.fill(self.compart_color, line_info)

    def load_piece_images(self, pieces: list):
        """
        Load the image files to the corresponding piece objects
        """
        for i in range(1, len(pieces)):
            pieces[i].set_basic_image()
            pieces[i].set_select_image()
            pieces[i].set_mini_image()

    def to_real_coor(self, clicked_coor):
        """
        Convert clicked coordinate to real coordinate
        """
        clicked_real_x = (clicked_coor[0] - COOR_OFFSET) // COOR_DELTA

        adj_y_offset = (COOR_OFFSET + BOARD_Y_OFFSET)
        clicked_real_y = (clicked_coor[1] - adj_y_offset) // COOR_DELTA

        return int(clicked_real_x), int(clicked_real_y)

    def find_target_piece(self, clicked_coor):
        """
        Search for the currently selected piece object
        # if the object is found, return True
        # else, return False
        """
        # get clicked coordinate
        clicked_x, clicked_y = clicked_coor

        # find the piece where the clicked coord is within its range
        for piece_id, piece in enumerate(self.ally_piece[1:], 1):
            if piece.state == DEAD:
                continue

            piece_x, piece_y = piece.get_pygame_coor()

            x_min = piece_x - piece.piece_width/2 + 25
            x_max = piece_x + piece.piece_width/2 + 25

            valid_x = x_min < clicked_x < x_max

            y_min = piece_y - piece.piece_height/2 + 25
            y_max = piece_y + piece.piece_height/2 + 25

            valid_y = y_min < clicked_y < y_max

            # if the clicked coord is within the piece range, select it
            if valid_x and valid_y:
                self.cur_selected = piece
                self.cur_selected_pid = piece_id
                break

    def init_timer(self):
        """
        Initialize the timer
        """
        self.time_font = pygame.font.SysFont('cochin', 20)
        self.timer_event = pygame.USEREVENT + 1
        pygame.time.set_timer(self.timer_event, 1000)

    def update_timer(self):
        """
        Update the remaining time and blit
        """
        timer_text = "Timer: " + str(self.counter)
        if self.counter <= 10:
            color = (230, 100, 100)  # Red
        else:
            color = (0, 0, 0)  # Black
        final_text = self.time_font.render(timer_text, True, color)
        text_rect = final_text.get_rect(centerx=460, bottom=720)
        self.screen.blit(final_text, text_rect)

    def init_kills(self):
        """
        Write 'Ally Kills: ' and 'Enemy Kills: ' on screen
        """
        kill_font = pygame.font.SysFont('cochin', 20)

        # Write Ally
        ally_text = "Ally"
        final_text = kill_font.render(ally_text, True, (20, 20, 0))
        text_rect = final_text.get_rect(centerx=55, bottom=740)
        self.screen.blit(final_text, text_rect)

        # Write Enemy
        enemy_text = "Enemy"
        final_text = kill_font.render(enemy_text, True, (20, 20, 0))
        text_rect = final_text.get_rect(centerx=55, bottom=50)
        self.screen.blit(final_text, text_rect)

        # Write Kills
        kill_text = " Kills"
        final_text = kill_font.render(kill_text, True, (20, 20, 0))

        text_rect = final_text.get_rect(centerx=55, bottom=770)
        self.screen.blit(final_text, text_rect)

        text_rect = final_text.get_rect(centerx=55, bottom=80)
        self.screen.blit(final_text, text_rect)

        # Draw border line
        border_font = pygame.font.SysFont('cochin', 30)
        border_text = "|"
        final_text = border_font.render(border_text, True, (20, 20, 0))
        text_rect = final_text.get_rect(centerx=90, bottom=755)
        self.screen.blit(final_text, text_rect)
        text_rect = final_text.get_rect(centerx=90, bottom=775)
        self.screen.blit(final_text, text_rect)
        text_rect = final_text.get_rect(centerx=90, bottom=65)
        self.screen.blit(final_text, text_rect)
        text_rect = final_text.get_rect(centerx=90, bottom=85)
        self.screen.blit(final_text, text_rect)

    def update_kills(self):
        """
        update the kills for both side
        """
        self.init_kills()

        self.ally_kills = ([enemy.mini_image for enemy in self.enemy_piece[1:]
                            if enemy.state == DEAD])
        self.enemy_kills = ([ally.mini_image for ally in self.ally_piece[1:]
                            if ally.state == DEAD])

    def render_kills(self):
        """
        This method renders capture logs for both sides.

        The coordinates (x,y) with in GUI are determined based on the number
        of current kills that the ally or the enemy side has during the game.

        These are implemented in two different loops because
        the number of dead pieces on both sides may differ during the game.

        x:
        - The x offsets 100 indicate the starting x coordinate on screen.

        - (i * 35) indicates the step size that considers both PIECE_WIDTH
            and proper spacing between the pieces to be listed.

        - Modulo 280 only allows the max number of pieces in each row to 8,
            therefore keeps the listed pieces within the pygame screen.

        y:
        - The y offsets 23 and 713 indicate the starting y coordinate
            for 'Enemy Kills' and 'Ally Kills' respectively on screen.

        - (i // 8) * 35 indicates that every piece in the position of
            multiple of 9 (ex] 9, 18), has to start a new line.
            Otherwise, the pieces will overlap and turn invisible.

        The modulo for y needed not to be set since we have enough spaces
        on the screen to handle the pieces even if they were all dead.
        """
        for i in range(len(self.ally_kills)):
            # keep minimis within the box
            x = 100 + (i * 35) % 280
            y = 713 + (i // 8) * 35
            self.screen.blit(self.ally_kills[i], (x, y))

        for i in range(len(self.enemy_kills)):
            # keep minimis within the box
            x = 100 + (i * 35) % 280
            y = 23 + (i // 8) * 35
            self.screen.blit(self.enemy_kills[i], (x, y))

    def kill_piece(self, real_clicked_coor):
        """
        Kill the enemy piece object in the given coordinate
        """
        for piece_id, enemy in enumerate(self.enemy_piece[1:], 1):
            if real_clicked_coor == enemy.coor and enemy.is_alive():
                enemy.state = DEAD
                break
        return piece_id

    def game_over(self):
        """
        Write the "game over" message on screen and wait for 3 seconds
        """
        game_over = "GAME OVER"
        font = pygame.font.SysFont('impact', 100)
        game_over_text = font.render(game_over, True, (128, 250, 128))
        t_rect = game_over_text.get_rect(center=self.screen.get_rect().center)
        self.screen.blit(game_over_text, t_rect)
        pygame.display.update()
        time.sleep(3)



import time

from gym_xiangqi.agents import RandomAgent
from gym_xiangqi.constants import (     # NOQA
    RED, BLACK, PIECE_ID_TO_NAME, ALLY
)
from gym_xiangqi.utils import action_space_to_move
from gym_xiangqi.envs import XiangQiEnv


def main():
    # Pass in the color you want to play as (RED or BLACK)
    env = XiangQiEnv(RED)
    env.render()
    agent = RandomAgent()

    done = False
    round = 0
    while not done:
        if env.turn == ALLY:
            _, reward, done, info = env.step_user()

            if "exit" in info and info["exit"]:
                break

            player = "You"
            piece, start, end = env.user_move_info
            piece = PIECE_ID_TO_NAME[piece]
        else:
            time.sleep(1)
            action = agent.move(env)
            _, reward, done, _ = env.step(action)

            player = "RL Agent"
            move = action_space_to_move(action)
            piece = PIECE_ID_TO_NAME[move[0]]
            start = move[1]
            end = move[2]

        env.render()
        round += 1
        print(f"Round: {round}")
        print(f"{player} made the move {piece} from {start} to {end}.")
        print(f"Reward: {reward}")
        print("================")

    print("Closing Xiangqi environment")
    env.close()


if __name__ == '__main__':
    main()



import os

import pkg_resources
import pygame

from gym_xiangqi.constants import BOARD_WIDTH, BOARD_HEIGHT, PATH_TO_BOARD


class Board:
    """
    A class for Xianqi Board
    """
    def __init__(self):
        self.boardWidth = BOARD_WIDTH
        self.boardHeight = BOARD_HEIGHT
        self.board_background = None
        self.load_board_image()

    def load_board_image(self):
        target_file = os.path.join(PATH_TO_BOARD, "BOARD.png")
        target_file = pkg_resources.resource_filename(__name__, target_file)
        self.board_background = pygame.image.load(target_file).convert_alpha()



import unittest

from gym_xiangqi.piece import (
    Piece, General, Advisor, Elephant,
    Horse, Chariot, Cannon, Soldier
)
from gym_xiangqi.constants import (
    ALLY, ENEMY,
    RED, BLACK, GENERAL, ADVISOR_1,
    HORSE_1, ELEPHANT_1, SOLDIER_1,
    CHARIOT_1, CANNON_1
)
from gym_xiangqi.utils import is_ally
from gym_xiangqi.envs.xiangqi_env import XiangQiEnv


class TestPieceClasses(unittest.TestCase):

    def setUp(self) -> None:
        self.classes = [Piece, General, Advisor, Elephant,
                        Horse, Chariot, Cannon, Soldier]

    def diff_move_list(self, env, piece_id, expected):
        """
        Helper function to sort then diff possible moves list of a piece.

        Parameters:
            env (XiangQiEnv): Environment used to diff the move list.
            piece_id (int): ID of the piece to test.
            expected (List[Tuple]): Expected moves from the piece in
                the format: [
                    (piece_id, [src_x, src_y], [dest_x, dest_y]),
                    (...)]
        """
        if is_ally(piece_id):
            pieces = env.ally_piece
            env.get_possible_actions(ALLY)
        else:
            pieces = env.enemy_piece
            env.get_possible_actions(ENEMY)

        env.get_possible_actions_by_piece(piece_id)
        result = pieces[abs(piece_id)].legal_moves
        result.sort()
        expected.sort()
        self.assertEqual(result, expected)

    def move_piece(self, env, piece_id, destination):
        """
        Simple helper function to move a piece to a destination.
        """
        source = (env.ally_piece[piece_id].row,
                  env.ally_piece[piece_id].col)
        env.ally_piece[piece_id].row = destination[0]
        env.ally_piece[piece_id].col = destination[1]
        env.state[destination[0]][destination[1]] = (
            env.state[source[0]][source[1]])
        env.state[source[0]][source[1]] = 0

    def test_piece_initialization(self):
        for piece_class in self.classes:
            piece = piece_class(0, 0, 0)
            self.assertIsInstance(piece, piece_class)
            self.assertEqual(piece.color, RED)
            self.assertEqual(piece.row, 0)
            self.assertEqual(piece.col, 0)
            piece.color = 1
            self.assertEqual(piece.color, BLACK)

    def test_general_can_move_within_palace(self):
        env = XiangQiEnv()
        # Test red general can only move forward 1 position from
        # starting position.
        self.diff_move_list(
            env=env,
            piece_id=GENERAL,
            expected=[([9, 4], [8, 4])]
        )

        # Move general to the center of the palace.
        self.move_piece(env=env,
                        piece_id=GENERAL,
                        destination=(8, 4))

        # Test again with general in the center of the palace.
        self.diff_move_list(
            env=env,
            piece_id=GENERAL,
            expected=[
                ([8, 4], [7, 4]),
                ([8, 4], [9, 4]),
                ([8, 4], [8, 5]),
                ([8, 4], [8, 3]),
            ]
        )

    def test_general_cannnot_move_out_of_palace(self):
        env = XiangQiEnv()

        # Move general to the top right hand corner of palace.
        self.move_piece(env=env,
                        piece_id=GENERAL,
                        destination=(7, 5))

        self.diff_move_list(
            env=env,
            piece_id=GENERAL,
            expected=[
                ([7, 5], [7, 4]),
                ([7, 5], [8, 5]),
            ]
        )

    def test_advisor_can_move_within_palace(self):
        env = XiangQiEnv()
        # Move the left advisor to the center of palace.
        self.move_piece(env=env,
                        piece_id=ADVISOR_1,
                        destination=(8, 4))

        self.diff_move_list(
            env=env,
            piece_id=ADVISOR_1,
            expected=[
                ([8, 4], [9, 3]),
                ([8, 4], [7, 3]),
                ([8, 4], [7, 5]),
            ]
        )

    def test_advisor_cannot_move_out_of_palace(self):
        env = XiangQiEnv()
        # Move the left advisor to the top left of palace.
        self.move_piece(env=env,
                        piece_id=ADVISOR_1,
                        destination=(7, 3))

        self.diff_move_list(
            env=env,
            piece_id=ADVISOR_1,
            expected=[
                ([7, 3], [8, 4]),
            ]
        )

    def test_horse_can_move(self):
        env = XiangQiEnv()
        # Move horse to an open space.
        self.move_piece(env=env,
                        piece_id=HORSE_1,
                        destination=(5, 4))

        self.diff_move_list(
            env=env,
            piece_id=HORSE_1,
            expected=[
                ([5, 4], [3, 5]),
                ([5, 4], [3, 3]),
                ([5, 4], [4, 6]),
                ([5, 4], [4, 2]),
            ]
        )

    def test_horse_will_be_blocked_by_pieces(self):
        env = XiangQiEnv()

        # Move horse to be in between 2 soldiers
        # and in front of cannon.
        self.move_piece(env=env,
                        piece_id=HORSE_1,
                        destination=(6, 1))

        self.diff_move_list(
            env=env,
            piece_id=HORSE_1,
            expected=[
                ([6, 1], [4, 0]),
                ([6, 1], [4, 2]),
            ]
        )

    def test_elephant_can_move(self):
        env = XiangQiEnv()

        self.diff_move_list(
            env=env,
            piece_id=ELEPHANT_1,
            expected=[
                ([9, 2], [7, 4]),
                ([9, 2], [7, 0]),
            ]
        )

    def test_elephant_cannot_cross_river(self):
        env = XiangQiEnv()

        # Move elephant to edge of river.
        self.move_piece(env=env,
                        piece_id=ELEPHANT_1,
                        destination=(5, 2))

        self.diff_move_list(
            env=env,
            piece_id=ELEPHANT_1,
            expected=[
                ([5, 2], [7, 4]),
                ([5, 2], [7, 0]),
            ]
        )

    def test_soldier_can_move(self):
        env = XiangQiEnv()

        self.diff_move_list(
            env=env,
            piece_id=SOLDIER_1,
            expected=[
                ([6, 0], [5, 0]),
            ]
        )

    def test_soldier_can_move_sideways_after_river(self):
        env = XiangQiEnv()

        # Move soldier across the river.
        self.move_piece(env=env,
                        piece_id=SOLDIER_1,
                        destination=(4, 0))

        self.diff_move_list(
            env=env,
            piece_id=SOLDIER_1,
            expected=[
                ([4, 0], [3, 0]),
                ([4, 0], [4, 1]),
            ]
        )

    def test_chariot_can_move(self):
        env = XiangQiEnv()

        # Move chariot to front of cannon.
        self.move_piece(env=env,
                        piece_id=CHARIOT_1,
                        destination=(6, 1))

        self.diff_move_list(
            env=env,
            piece_id=CHARIOT_1,
            expected=[
                ([6, 1], [5, 1]),
                ([6, 1], [4, 1]),
                ([6, 1], [3, 1]),
                ([6, 1], [2, 1]),
            ]
        )

    def test_cannon_can_skip_enemy_piece(self):
        env = XiangQiEnv()

        # Move cannon forward 1 position and put
        # chariot behind to limit possible moves.
        self.move_piece(env=env,
                        piece_id=CANNON_1,
                        destination=(6, 1))
        self.move_piece(env=env,
                        piece_id=CHARIOT_1,
                        destination=(7, 1))

        self.diff_move_list(
            env=env,
            piece_id=CANNON_1,
            expected=[
                ([6, 1], [5, 1]),
                ([6, 1], [4, 1]),
                ([6, 1], [3, 1]),
                ([6, 1], [0, 1]),
            ]
        )

    def test_cannon_can_skip_ally_piece(self):
        env = XiangQiEnv()

        # Move cannon and put horse in front of it.
        self.move_piece(env=env,
                        piece_id=CANNON_1,
                        destination=(6, 1))
        self.move_piece(env=env,
                        piece_id=HORSE_1,
                        destination=(5, 1))

        self.diff_move_list(
            env=env,
            piece_id=CANNON_1,
            expected=[
                ([6, 1], [2, 1]),
                ([6, 1], [7, 1]),
                ([6, 1], [8, 1]),
                ([6, 1], [9, 1]),
            ]
        )


if __name__ == "__main__":
    unittest.main()



import numpy as np
import random

from gym_xiangqi.constants import ALLY


class RandomAgent:
    """
    This is the implementation of the simplest
    agent possible to play the game of Xiang Qi.
    The agent will simply choose a random move
    out of all the possible moves and return that.
    """
    def __init__(self):
        pass

    def move(self, env):
        """
        Make a random move based on the environment.
        """
        actions = (env.ally_actions if env.turn == ALLY
                   else env.enemy_actions)
        legal_moves = np.where(actions == 1)[0]
        ind = random.randint(0, len(legal_moves)-1)
        return legal_moves[ind]



import unittest
from unittest.mock import patch
import random
import string

from gym_xiangqi.envs.xiangqi_env import XiangQiEnv
from gym_xiangqi.xiangqi_game import XiangQiGame
from gym_xiangqi.constants import (
    BOARD_ROWS, BOARD_COLS,
    RED, BLACK, DEAD,
    ILLEGAL_MOVE, PIECE_POINTS, LOSE,
    EMPTY, GENERAL, CANNON_1, HORSE_2,
    ALLY, ENEMY,
    INITIAL_BOARD,
)


class TestXiangQiEnv(unittest.TestCase):

    def assertStateEqual(self, obs, new_obs):
        for i in range(BOARD_ROWS):
            for j in range(BOARD_COLS):
                self.assertEqual(obs[i][j], new_obs[i][j])

    def setUp(self):
        self.env = XiangQiEnv()

    def tearDown(self):
        self.env.close()

    def test_env_initialization(self):
        self.assertEqual(self.env.ally_color, RED)
        self.assertEqual(self.env.enemy_color, BLACK)

        self.env = XiangQiEnv(ally_color=BLACK)
        self.assertEqual(self.env.ally_color, BLACK)
        self.assertEqual(self.env.enemy_color, RED)

        self.assertIsNotNone(self.env.game)
        self.assertIsNotNone(self.env.state)
        self.assertIsNotNone(self.env.observation_space)
        self.assertIsNotNone(self.env.action_space)
        self.assertIsNotNone(self.env.ally_actions)
        self.assertIsNotNone(self.env.enemy_actions)

    def test_env_step_invalid_action(self):
        """
        verify action input
        """
        with self.assertRaises(AssertionError):
            # Test using a random negative int.
            action = random.randint(-10000, -1)
            self.env.step(action)

            # Test using a random float.
            action = random.uniform(-10000, 10000)
            self.env.step(action)

            # Test using a random string.
            length = random.randint(0, 15)
            choices = (string.ascii_letters +
                       string.punctuation +
                       string.digits)
            action = "".join(random.choice(choices) for i in range(length))
            self.env.step(action)

    def test_flying_general(self):
        """
        Verification of flying general condition checker
        Given an action that results in flying general, the environment has
        to reject and penalize the agent
        """
        # pre-defined actions tht will result in a flying general situation
        actions = [78727, 75172, 78961, 74938, 75720, 78177]

        for action in actions:
            obs = self.env.state
            new_obs, reward, _, _ = self.env.step(action)

        # check that our state is preserved and the reward is a penalty
        self.assertStateEqual(obs, new_obs)
        self.assertEqual(reward, ILLEGAL_MOVE)

    def test_env_step_illegal_move(self):
        """
        verify agent with illegal move
        """
        obs = self.env.state
        new_obs, reward, done, info = self.env.step(0)

        self.assertStateEqual(obs, new_obs)
        self.assertEqual(reward, ILLEGAL_MOVE)
        self.assertEqual(done, False)

    def test_env_step_one_round(self):
        """
        verify one round of ally and enemy turns
        78727: Ally CANNON_1 (7, 1) -> (7, 4)
        75172: Enemy CANNON_1 (2, 7) -> (2, 4)
        """
        actions = [(78727, CANNON_1, (7, 1), (7, 4), 0, ENEMY),
                   (75172, CANNON_1, (2, 7), (2, 4), 0, ALLY)]

        for action, pid, (r1, c1), (r2, c2), points, next_turn in actions:
            obs, reward, done, info = self.env.step(action)
            self.assertEqual(obs[r1][c1], EMPTY)
            self.assertEqual(obs[r2][c2], pid * self.env.turn * -1)
            self.assertEqual(reward, points)
            self.assertEqual(done, False)
            self.assertEqual(self.env.turn, next_turn)

    def test_env_step_reward(self):
        """
        Simulates Ally (red) cannon taking a black horse
        78661: Ally CANNON_1 (7, 1) -> (0, 1)
        """
        action = 78661
        obs, reward, done, info = self.env.step(action)
        self.assertEqual(obs[7][1], EMPTY)
        self.assertEqual(obs[0][1], CANNON_1)
        self.assertEqual(reward, PIECE_POINTS[HORSE_2])
        self.assertEqual(done, False)
        self.assertEqual(self.env.enemy_piece[HORSE_2].state, DEAD)

    def test_env_step_done(self):
        """
        verify environment termination due to death of black general
        78727: Ally CANNON_1 (7, 1) -> (7, 4)
        75172: Enemy CANNON_1 (2, 7) -> (2, 4)
        78961: Ally CANNON_1 (7, 4) -> (3, 4)
        123966: Enemy SOLDIER_5 (3, 0) -> (4, 0)
        75694: Ally CANNON_1 (3, 4) -> (0, 4) -- takes black general
        """
        actions = [78727, 75172, 78961, 123966, 75694]
        for action in actions:
            obs, reward, done, info = self.env.step(action)
        self.assertEqual(done, True)
        self.assertEqual(reward, PIECE_POINTS[GENERAL])
        self.assertEqual(self.env.enemy_piece[GENERAL].state, DEAD)

    def test_env_reset(self):
        """
        verify environment properly resets after an episode has terminated

        Performs these series of actions:
        78727: Ally CANNON_1 (7, 1) -> (7, 4)
        75172: Enemy CANNON_1 (2, 7) -> (2, 4)
        78961: Ally CANNON_1 (7, 4) -> (3, 4)
        123966: Enemy SOLDIER_5 (3, 0) -> (4, 0)
        75694: Ally CANNON_1 (3, 4) -> (0, 4) -- takes black general

        and resets the environment.
        """
        actions = [78727, 75172, 78961, 123966, 75694]
        for action in actions:
            obs, reward, done, info = self.env.step(action)
        self.assertEqual(done, True)
        self.assertEqual(reward, PIECE_POINTS[GENERAL])
        self.assertEqual(self.env.enemy_piece[GENERAL].state, DEAD)

        obs = self.env.reset()
        self.assertFalse(self.env._done)
        self.assertStateEqual(INITIAL_BOARD, obs)

    def test_perpetual_check(self):
        """
        simulate a perpetual checking situation
        64062:      Chariot 1 (red)     (9, 0) -> (8, 0)
        57437:      Chariot 1 (black)   (0, 8) -> (1, 8)
        63255:      Chariot 1 (red)     (8, 0) -> (8, 3)
        373:        General (black)     (0, 4) -> (1, 4)
        63462:      Chariot 1 (red)     (8, 3) -> (1, 3)
        --------------- REPETITION START ---------------
        1192:       General (black)     (1, 4) -> (2, 4)
        57801:      Chariot 1 (red)     (1, 3) -> (2, 3)
        1993:       General (black)     (2, 4) -> (1, 4)
        58602:      Chariot 1 (red)     (2, 3) -> (1, 3)
        ---------------- REPETITION END ----------------
        """
        for action in [64062, 57437, 63255, 373, 63462]:
            obs, reward, done, _ = self.env.step(action)
            self.assertFalse(done)

        for _ in range(3):
            for action in [1192, 57801, 1993, 58602]:
                _, reward, done, _ = self.env.step(action)
        self.assertEqual(reward, LOSE)
        self.assertTrue(done)

    def test_env_close(self):
        self.env.render()
        self.env.close()

    def test_env_seed(self):
        seed_list = self.env.seed()
        self.assertIsNotNone(seed_list)
        self.assertIsInstance(seed_list, list)
        self.assertIsInstance(seed_list[0], int)

    def test_env_state_hash_check(self):
        self.env._state[0][7], self.env._state[2][1] = (
            self.env._state[2][1], self.env._state[0][7])
        with self.assertRaises(AssertionError):
            self.env.step(300)

    def test_env_step_user(self):
        def mock_run(this):
            this.cur_selected_pid = CANNON_1
            this.end_pos = (0, 1)

        with patch.object(XiangQiGame, 'run', new=mock_run):
            obs, reward, done, _ = self.env.step_user()

        self.assertEqual(self.env.user_move_info, (CANNON_1, (7, 1), (0, 1)))
        self.assertEqual(obs[7][1], EMPTY)
        self.assertEqual(obs[0][1], CANNON_1)
        self.assertEqual(reward, PIECE_POINTS[HORSE_2])
        self.assertFalse(done)


if __name__ == "__main__":
    unittest.main()



import os

import pkg_resources
import pygame

from gym_xiangqi.constants import PATH_TO_SOUNDS


class Sound:
    """
    A class for sounds
    """
    def __init__(self, piece_move, bgm):
        # declare variables
        self.piece_move = None

        # init pygame mixer
        pygame.init()
        pygame.mixer.init()

        # load sounds on init
        self.load_bgm(bgm)
        self.load_piece_move(piece_move)

    # load background music
    def load_bgm(self, _bgm):
        target_file = os.path.join(PATH_TO_SOUNDS, _bgm)
        target_file = pkg_resources.resource_filename(__name__, target_file)
        pygame.mixer.music.load(target_file)

    # load the sound effect for piece movements
    def load_piece_move(self, _piece_move):
        target_file = os.path.join(PATH_TO_SOUNDS, _piece_move)
        target_file = pkg_resources.resource_filename(__name__, target_file)
        self.piece_move = pygame.mixer.Sound(target_file)



from gym_xiangqi.envs.xiangqi_env import XiangQiEnv  # NOQA
from gym_xiangqi.agents import RandomAgent  # NOQA
import timeit

""" Timing """
# Number of times to repeat the timing tests.
NUM_REPEAT = 10
# Number of times to call a method for each timing test.
NUM_RUN = 1

""" Setup """
ENV_SETUP = "env = XiangQiEnv()"
AGENT_SETUP = "agent = RandomAgent()"


def print_time(title, time_list):
    """
    Helper function to print latency times
    in a nice format.
    """
    assert len(time_list) == NUM_REPEAT
    print(f"{title}")
    print("--------------------")
    print(f"min    : {min(time_list):.3f}")
    print(f"max    : {max(time_list):.3f}")
    print(f"average: {sum(time_list) / len(time_list):.3f}")
    print("\n")


def s_to_ms(time_list):
    return [t * 1000 for t in time_list]


def measure_time(cmd_str, setup=""):
    return s_to_ms(timeit.repeat(
        cmd_str, setup=setup, globals=globals(),
        repeat=NUM_REPEAT, number=NUM_RUN))


def measure_and_print_latency(methods_to_setup):
    """
    Helper function to measure and print the latency
    of methods specified in methods_to_setup.

    Parameters:
        methods_to_setup (dict[str, str]): Dictionary matching method calls
            to setups needed to perform the call.
    """
    for method, setup in methods_to_setup.items():
        time_list = measure_time(cmd_str=method, setup=setup)
        print_time(title=method, time_list=time_list)


def env_latency():
    """
    Measure and print the latency of the methods defined
    in XiangQiEnv.
    """
    print("XiangQiEnv Latency (ms)")
    print("=========================")

    methods_to_setup = {
        "XiangQiEnv()": "",
        "env.init_pieces()": ENV_SETUP,
        "env.get_possible_actions(player)": f"{ENV_SETUP}; player=1;",
        "env.step(action)":
            f"{ENV_SETUP}; import numpy as np; import random;"
            "actions = np.where(env.ally_actions == 1)[0];"
            "action = random.randint(0, len(actions)-1);"
    }
    measure_and_print_latency(methods_to_setup)


def random_agent_latency():
    """
    Measure and print the latency of the methods defined
    in RandomAgent.
    """
    print("RandomAgent Latency (ms)")
    print("=========================")

    methods_to_setup = {
        "RandomAgent()": "",
        "agent.move(env)": f"{ENV_SETUP}; {AGENT_SETUP}",
    }
    measure_and_print_latency(methods_to_setup)


if __name__ == "__main__":
    # TODO: Compare latency from before and after a PR, and actually
    # fail a PR if the change if too big or if the latency
    # exceed a certain number.
    env_latency()
    random_agent_latency()
